package geniusweb.protocol.session.saop;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.mockito.Matchers.any;
import static org.mockito.Matchers.anyLong;
import static org.mockito.Matchers.eq;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import java.io.IOException;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import org.junit.Test;
import org.mockito.ArgumentCaptor;

import geniusweb.actions.Accept;
import geniusweb.actions.Action;
import geniusweb.actions.EndNegotiation;
import geniusweb.actions.Offer;
import geniusweb.actions.PartyId;
import geniusweb.deadline.Deadline;
import geniusweb.deadline.DeadlineTime;
import geniusweb.events.CurrentState;
import geniusweb.events.ProtocolEvent;
import geniusweb.party.inform.Finished;
import geniusweb.party.inform.Inform;
import geniusweb.party.inform.Settings;
import geniusweb.party.inform.YourTurn;
import geniusweb.progress.Progress;
import geniusweb.protocol.ProtocolException;
import geniusweb.protocol.partyconnection.ProtocolToPartyConn;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnFactory;
import geniusweb.protocol.partyconnection.ProtocolToPartyConnections;
import geniusweb.protocol.session.TeamOfPartiesAndProfiles;
import geniusweb.references.Parameters;
import geniusweb.references.PartyRef;
import geniusweb.references.PartyWithParameters;
import geniusweb.references.PartyWithProfile;
import geniusweb.references.ProfileRef;
import geniusweb.references.ProtocolRef;
import tudelft.utilities.listener.Listener;
import tudelft.utilities.logging.ReportToLogger;
import tudelft.utilities.repository.NoResourcesNowException;

/**
 * This test tests inner workings of the SAOP protocol, AKA 'white box' testing.
 * This adds brittleness to this test, because these internal workings may be
 * modified as long as the public API remains working. This approach was
 * necessary because fully testing the public API directly would lead to
 * excessive amounts of mocking and unfocused tests. (with "focused test" we
 * mean that a junit test should test a small part of the code and thus aid in
 * locating the actual issue).
 */
public class SAOPTest {
	private final SAOPState state = mock(SAOPState.class),
			connectedstate = mock(SAOPState.class),
			failstate = mock(SAOPState.class),
			finalstate = mock(SAOPState.class);

	private final ProtocolRef SAOP;
	private static final PartyId PARTY2ID = new PartyId("party2");
	private static final PartyId PARTY1ID = new PartyId("party1");
	private final PartyRef party1ref = mock(PartyRef.class);
	private final PartyRef party2ref = mock(PartyRef.class);
	private final SAOPSettings settings = mock(SAOPSettings.class);
	private final SaopPartyWithProfile team1 = mock(SaopPartyWithProfile.class);
	private final SaopPartyWithProfile team2 = mock(SaopPartyWithProfile.class);
	private SAOP saop;
	private ProtocolToPartyConnFactory factory;
	private List connections;
	private ProtocolToPartyConn conn1 = mock(ProtocolToPartyConn.class),
			conn2 = mock(ProtocolToPartyConn.class);
	private Map<PartyId, SaopPartyWithProfile> partyprofiles;
	private Progress progress = mock(Progress.class);
	private ProfileRef profile1;
	private ProfileRef profile2;
	private ProtocolToPartyConnections connectionswithparties = mock(
			ProtocolToPartyConnections.class);;
	@SuppressWarnings("unchecked")
	private final Listener<ProtocolEvent> testlistener = mock(Listener.class);
	private final Parameters parameters = new Parameters();

	private final long NOW = 1000;
	private final PartyWithParameters partywithparam1 = new PartyWithParameters(
			party1ref, parameters);
	private final PartyWithParameters partywithparam2 = new PartyWithParameters(
			party2ref, parameters);

	private final Deadline deadlinetime = mock(DeadlineTime.class);

	public SAOPTest()
			throws URISyntaxException, IOException, NoResourcesNowException {
		SAOP = new ProtocolRef("SAOP");
		when(team1.getParty()).thenReturn(partywithparam1);
		when(team2.getParty()).thenReturn(partywithparam2);
		when(team1.getAllParties()).thenReturn(Arrays.asList(team1));
		when(team2.getAllParties()).thenReturn(Arrays.asList(team2));

		partyprofiles = new HashMap<>();
		partyprofiles.put(PARTY1ID, team1);
		partyprofiles.put(PARTY2ID, team2);

		when(deadlinetime.getDuration()).thenReturn(1000l);

		List<TeamOfPartiesAndProfiles> teams = new ArrayList<>();
		teams.add(team1);
		teams.add(team2);
		when(settings.getTeams()).thenReturn(teams);
		when(settings.getAllParties()).thenReturn(Arrays.asList(team1, team2));
		when(settings.getDeadline()).thenReturn(deadlinetime);

		factory = mock(ProtocolToPartyConnFactory.class);
		connections = mock(List.class);
		when(factory.connect(any(List.class))).thenReturn(connections);

		// hack the state.with(connection) stuff simplistically

		when(conn1.getParty()).thenReturn(PARTY1ID);
		when(conn2.getParty()).thenReturn(PARTY2ID);

		profile1 = mock(ProfileRef.class);
		when(team1.getProfile()).thenReturn(profile1);
		profile2 = mock(ProfileRef.class);
		when(team2.getProfile()).thenReturn(profile2);

		mockState(connectedstate, "connected state");
		mockState(state, "initial state");
		mockState(finalstate, "final state");
		when(finalstate.isFinal(anyLong())).thenReturn(true);
		mockState(failstate, "fail state");
		when(failstate.isFinal(anyLong())).thenReturn(true);

		when(connectionswithparties.get(eq(PARTY1ID))).thenReturn(conn1);
		when(connectionswithparties.get(eq(PARTY2ID))).thenReturn(conn2);
		// HACK thenReturn twice, so that 2nd call to iterator() returns new
		// iterator instead of the old one
		when(connectionswithparties.iterator())
				.thenReturn(Arrays.asList(conn1, conn2).iterator())
				.thenReturn(Arrays.asList(conn1, conn2).iterator());
		saop = new SAOP(state, new ReportToLogger("test"));
		saop.addListener(testlistener);
	}

	/**
	 * All states are more or less the same, but are slightly modified
	 * 
	 * @param state  the state to mock (must be already
	 * @param asText
	 */
	private void mockState(SAOPState state, String asText) {
		when(state.getSettings()).thenReturn(settings);
		when(state.getConnections()).thenReturn(connectionswithparties);
		when(state.getPartyProfiles())
				.thenReturn(partyprofiles.entrySet().stream().collect(
						Collectors.toMap(e -> e.getKey(), e -> e.getValue())));
		when(state.getProgress()).thenReturn(progress);
		when(state.with(any(ProtocolToPartyConn.class),
				any(PartyWithProfile.class))).thenReturn(connectedstate);
		when(state.getNextActor()).thenReturn(PARTY1ID);
		when(state.with(any(PartyId.class), any(EndNegotiation.class)))
				.thenReturn(finalstate);
		when(state.with(any(PartyId.class), any(Accept.class)))
				.thenReturn(state);
		when(state.with(any(Progress.class))).thenReturn(state);
		when(state.toString()).thenReturn(asText);
		when(state.with(any(ProtocolException.class))).thenReturn(failstate);
	}

	@Test
	public void testConstructor() {
		SAOPState state1 = mock(SAOPState.class);
		Deadline dl = mock(Deadline.class);
		SAOPSettings set = mock(SAOPSettings.class);
		when(state1.getSettings()).thenReturn(set);
		when(set.getDeadline()).thenReturn(dl);
		when(dl.getDuration()).thenReturn(1000l);
		new SAOP(state1, new ReportToLogger("test"));
		verify(state1, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testConnect()
			throws InterruptedException, IOException, NoResourcesNowException {
		saop.connect(factory);

		assertEquals(connectedstate, saop.getState());
		verify(state, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));
	}

	@Test(expected = IOException.class)
	public void testConnectFailingConnection()
			throws InterruptedException, IOException, NoResourcesNowException {
		// override the factory connect to throw IOException. This exception
		// should boil up and cause the connect to fail.
		when(factory.connect(any(List.class)))
				.thenThrow(new IOException("Refusing connection"));
		saop.connect(factory);
		verify(state, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testConnectRetry()
			throws InterruptedException, IOException, NoResourcesNowException {
		// override the factory connect to throw NoResourcesNowException and
		// then succeed 2nd time.
		when(factory.connect(any(List.class)))
				.thenThrow(new NoResourcesNowException("Refusing connection",
						new Date(System.currentTimeMillis() + 500)))
				.thenReturn(connections);
		saop.connect(factory);

		assertEquals(connectedstate, saop.getState());
		verify(state, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testSetup() throws ProtocolException, IOException {

		saop.setupParties();

		// were the connections attached properly?
		verify(conn1, times(1)).addListener(any(Listener.class));
		verify(conn2, times(1)).addListener(any(Listener.class));

		// were the settings send to each party?
		verify(conn1, times(1)).send(any(Settings.class));
		ArgumentCaptor<Settings> actualsettings = ArgumentCaptor
				.forClass(Settings.class);
		verify(conn1).send(actualsettings.capture());
		assertEquals(
				new Settings(PARTY1ID, profile1, SAOP, progress, parameters),
				actualsettings.getValue());

		verify(conn2, times(1)).send(any(Settings.class));
		verify(conn2).send(actualsettings.capture());
		assertEquals(
				new Settings(PARTY2ID, profile2, SAOP, progress, parameters),
				actualsettings.getValue());
		verify(state, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));

	}

	@Test
	public void testActionRequestWrongActor() {

		saop.actionRequest(conn2, mock(EndNegotiation.class));

		verify(state, times(1)).with(any(ProtocolException.class));
		verify(testlistener, times(1)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testActionRequest() throws IOException {
		when(state.getNextActor()).thenReturn(PARTY1ID);
		when(state.with(any(PartyId.class), any(Action.class)))
				.thenReturn(state);

		saop.actionRequest(conn1, mock(EndNegotiation.class));

		verify(state, times(0)).with(any(ProtocolException.class));
		verify(state, times(1)).with(eq(PARTY1ID), any(EndNegotiation.class));
		verify(connectionswithparties, times(1)).broadcast(any(Inform.class));

		// state.getNextActor() is frozen to PARTY1 by our mocking
		verify(conn1, times(1)).send(any(YourTurn.class));
		verify(conn2, times(0)).send(any(YourTurn.class));
		verify(testlistener, times(0)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testFinalActionRequest() throws IOException {
		when(state.getNextActor()).thenReturn(PARTY1ID);
		when(state.with(any(PartyId.class), any(EndNegotiation.class)))
				.thenReturn(finalstate);

		saop.actionRequest(conn1, mock(EndNegotiation.class));

		verify(state, times(0)).with(any(ProtocolException.class));
		verify(state, times(1)).with(eq(PARTY1ID), any(EndNegotiation.class));
		verify(connectionswithparties, times(1)).broadcast(any(Finished.class));
		verify(testlistener, times(1)).notifyChange(any(CurrentState.class));

		verify(conn1, times(1)).send(any(YourTurn.class));
		verify(conn2, times(1)).send(any(YourTurn.class));
	}

	@Test
	public void testStart() {
		saop.start(factory);
	}

	@Test
	public void testActionFailNextYourturn() throws IOException {
		doThrow(new IOException("fail sending yourturn")).when(conn1)
				.send(any(YourTurn.class));

		saop.actionRequest(conn1, mock(Accept.class));

		verify(state, times(1)).with(any(ProtocolException.class));
		verify(testlistener, times(1)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testActionNotTheTurn() throws IOException {
		doThrow(new IOException("fail sending yourturn")).when(conn1)
				.send(any(YourTurn.class));
		// not turn of conn2.
		saop.actionRequest(conn2, mock(EndNegotiation.class));

		verify(state, times(1)).with(any(ProtocolException.class));
		verify(testlistener, times(1)).notifyChange(any(CurrentState.class));
	}

	@Test
	public void testActionInFinalState() throws IOException {
		saop = new SAOP(finalstate, new ReportToLogger("test"));

		saop.actionRequest(conn1, mock(Offer.class));

		verify(finalstate, times(0)).with(any(ProtocolException.class));
		verify(testlistener, times(0)).notifyChange(any());
	}

	@Test
	public void testDescription() {
		assertNotNull(saop.getDescription());
	}

	@Test(expected = IllegalStateException.class)
	public void testAddParticipant() {
		saop.addParticipant(null);
	}
}
